""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import ast
import json
import os
import uuid
import requests
import hashlib
import logging
from django.conf import settings
from cshmac.requests import HmacAuth
from .requests_auth import get_requests_auth
from requests_toolbelt.downloadutils import stream
from urllib.parse import urlparse
from collections import namedtuple
from django.http.multipartparser import parse_header

logger = logging.getLogger('connectors')

# File handling
FileMetadata = namedtuple('FileMetadata', ['filename',
                                           'content_length',
                                           'content_type',
                                           'md5',
                                           'sha1',
                                           'sha256'])


def _get_request_params(method, body=None, *args, **kwargs):
    env = kwargs.get('env', {})
    bodyless_methods = ['head', 'get']
    file_operation = kwargs.get('file_operation', False)

    if not isinstance(env, dict):
        env = {}
    # for default if no env HMAC method
    if not env or not env.get('auth_info', False):
        env['auth_info'] = {"auth_method": "CS HMAC"}

    # for public and private key in env
    if env.get('public_key', False) and env.get('private_key', False):
        env['auth_info'] = {"auth_method": "CS HMAC"}
        public_key = env.get('public_key')
        private_key = env.get('private_key')
    else:
        public_key = settings.APPLIANCE_PUBLIC_KEY
        private_key = settings.APPLIANCE_PRIVATE_KEY
    auth_info = env.get('auth_info')

    if not body and not method.lower() in bodyless_methods:
        body = public_key if file_operation else {}

    return body, auth_info, public_key, private_key


def make_request(url, method, body=None, *args, **kwargs):
    """
    This function facilitates using the crud hub api.

    It is for general purpose requests, but takes care of authentication
    automatically.

   :param str collection: An IRI that points to the location of the \
       crud hub collection (E.g. /api/3/events)
   :param str method: HTTP method
   :param dict body: An object to json encode and send to crud hub

   :return: the API response either as a json-like dict if possible or as bytes
   :rtype: dict/bytes
    """
    # ctrl c/v
    # get rid of the body on GET/HEAD requests

    # account for relative iri
    request_param_keys = ['params', 'data', 'headers', 'cookies', 'files', 'timeout', 'allow_redirects', 'proxies',
                          'stream', 'cert']
    request_params = {}

    if not bool(urlparse(url).netloc):
        url = settings.CRUD_HUB_URL + str(url)
    else:
        url = url

    bodyless_methods = ['head', 'get']
    if method.lower() in bodyless_methods:
        body = None

    if type(body) == str:
        try:
            body = ast.literal_eval(body)
        except Exception:
            pass

    logger.info('Starting request: %s , %s', method, url)

    for param_key in request_param_keys:
        param_value = kwargs.get(param_key)
        if param_value or isinstance(param_key, bool):
            request_params[param_key] = param_value

    body, auth_info, public_key, private_key = _get_request_params(method, body=body, *args, **kwargs)
    auth = get_requests_auth(auth_info, url, method,
                             public_key,
                             private_key,
                             json.dumps(body), *args, **kwargs)

    response = requests.request(method, url, auth=auth, json=body, verify=False, **request_params)
    if not kwargs.get('validate_response', True):
        return response
    return maybe_json_or_raise(response)

def trigger_ingest_playbook(records, playbook_uuid, playbook_input_param = "ingestedData", parent_env={}, batch_size=1000, dedup_field=None):
    if dedup_field:
        seen = set()
        records = [x for x in records if [x[dedup_field].replace(" ", "") not in seen, seen.add(x[dedup_field].replace(" ", ""))][0]]
    url = "/api/triggers/1/notrigger/" + playbook_uuid
    method = "POST"
    
    parent_wf = parent_env.get('wf_id')
    parent_step = parent_env.get('step_id')
    try:
        for start_index in range(0, len(records), batch_size):
            payload = {
            "priority": {"itemValue": "Low"},
            "_eval_input_params_from_env": True,
            "env": {playbook_input_param: records[start_index: start_index + batch_size]}
            } 
            if parent_wf:
                payload['parent_wf'] = parent_wf
            if parent_step:
                payload['step_id'] = parent_step
            make_request(url, method, body=payload)
    except Exception as e:
        logger.error("Failed to insert a batch of feeds with error: " + str(e))

def make_file_upload_request(file_name, file_content, file_type, collection='files', *args, **kwargs):
    multipart_headers = kwargs.get('multipart_headers') or {'Expire': 0}
    request_headers = kwargs.get('request_headers') or {}
    extra_fields = kwargs.get('extra_multipart_fields') or {}
    url = settings.CRUD_HUB_URL + '/api/3/{0}'.format(collection)
    body, auth_info, public_key, private_key = _get_request_params("POST", file_operation=True, *args, **kwargs)
    auth = get_requests_auth(auth_info, url, "POST",
                             public_key,
                             private_key,
                             body, *args, **kwargs)
    files = {'file': (file_name, file_content, file_type, multipart_headers)}
    files.update(extra_fields)
    response = requests.post(url, auth=auth, files=files, verify=False, headers=request_headers)
    return maybe_json_or_raise(response)


def maybe_json_or_raise(response):
    """
    Helper function for processing request responses

    Returns any json found in the response. Otherwise, it will extract the
    response as text, or, failing that, as bytes.

    :return: the response from the request
    :rtype: dict or str or bytes
    :raises: :class:`requests.HTTPError` if status code was 4xx or 5xx
    """
    if response.ok:
        try:
            logger.info('Processing request responses.')
            return response.json(strict=False)
        except Exception:
            return response.text or response.content
    else:
        msg = ''
        try:
            msg = response.json()
            logger.warn(msg)
        except Exception:
            pass
        if not msg:
            msg = response.text
            logger.warn(msg)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # add any response content to the error message
            msg = '{} :: {} :: Url: {}'.format(str(e), msg, response.url)
            logger.error(msg)
            raise requests.exceptions.HTTPError(msg, response=response)


def download_file_from_cyops(iri, headers=None, *args, **kwargs):
    if not bool(urlparse(iri).netloc):
        iri = settings.CRUD_HUB_URL + str(iri)

    # get the @id and type (with or without a trailing slash)
    split = iri.split('/')
    obj_id = split[-1]
    if not obj_id:
        obj_id = split[-2]
        obj_type = split[-3]
    else:
        obj_type = split[-2]

    # retrieve the file info from the attachment
    if obj_type != 'files':
        collection = '/api/3/{obj_type}/{obj_id}'.format(obj_type=obj_type,
                                                         obj_id=obj_id)
        logger.info('Download file from crudhub %s', collection)
        res = make_request(collection, 'GET', None, *args, **kwargs)
        file_iri = res['file']

        if type(file_iri) is not str:
            file_iri = file_iri.get('@id', '')
        return download_file_from_cyops(file_iri, *args, **kwargs)

    collection = '/api/3/files?uuid={}'.format(obj_id)
    logger.info('Download file metadata %s', collection)
    metadata = make_request(collection, 'GET', None, *args, **kwargs)
    ch_metadata = metadata['hydra:member'][0]
    mime_type = ch_metadata.get('mimeType', 'application/octet-stream')
    if not headers:
        headers = {}
    headers.update({'Accept': 'application/octet-stream'})
    # download the file
    filename, meta = download_file(iri, is_crudhub_endpoint=True, headers=headers, *args, **kwargs)
    try:
        metadata = FileMetadata(ch_metadata.get('filename', filename),
                                ch_metadata.get('size', 0),
                                ch_metadata.get('mimeType', ''),
                                ch_metadata.get('md5', meta.md5),
                                ch_metadata.get('sha1', meta.sha1),
                                ch_metadata.get('sha256', meta.sha256))
    except Exception as e:
        logger.error(str(e))
        metadata = FileMetadata(filename, 0, '', '', '', '')
        # save in env
        save_file_in_env(kwargs.get('env', {}), filename, metadata)
    # return the voucher
    res = metadata._asdict()
    res.update({'cyops_file_path': filename})
    save_file_in_env(kwargs.get('env', {}), filename, metadata)
    return res


def download_file(iri, auth=None, headers=None, *args, **kwargs):
    logger.info('download_file from arbitrary iri over http IRI: %s', iri)
    if headers is None:
        headers = {}

    request_args = {
        'stream': True,
        'verify': False,
        'headers': headers,
    }
    # see body-content-workflow at
    # http://docs.python-requests.org/en/master/user/advanced/
    if kwargs.get('is_crudhub_endpoint'):
        request_args['method'] = 'GET'
        request_args['url'] = iri
        request_args['validate_response'] = False
        response = make_request(**request_args)
    else:
        request_args['url'] = iri
        request_args['auth'] = auth
        response = requests.get(**request_args)
    # error handling first
    if not response.ok:
        result = {
            'status_code': response.status_code,
            'message': response.content,
        }
        raise Exception(result)

    # consume the response
    file_path = read_response_to_file(response)

    # get filename from the absolute file path
    filename = os.path.basename(file_path)

    content_disposition = response.headers.get('content-disposition', '')
    try:
        # parse_headers expects a latin-1 header byte string
        content_disposition = content_disposition.encode('iso-8859-1')
    except AttributeError:
        logger.error('Error occurred while parsing the headers received by request call')

    # e.g. ('conntent-disposition: attachment', {'filename': b'\xad'})
    key, filename_dict = parse_header(content_disposition)

    # get and convert filename from the filename dict
    cd_filename = filename_dict.get('filename', None)
    try:
        content_disposition_filename = cd_filename.decode('utf-8')
    except AttributeError:
        content_disposition_filename = cd_filename

    # if that didn't work, guess the filename from the url
    if not content_disposition_filename:
        content_disposition_filename = response.url.split('/')[-1]

    # if *that* didn't work, use the generated uuid
    content_disposition_filename = content_disposition_filename or filename

    # extract other header metadata, if available
    file_size = response.headers.get('content-length', 'unknown')
    content_type = response.headers.get('content-type', 'unknown')
    file_hashes = calculate_hashes(file_path)
    # construct full file metadata tuple
    metadata = FileMetadata(content_disposition_filename,
                            file_size, content_type, file_hashes['md5'],
                            file_hashes['sha1'], file_hashes['sha256'])

    return filename, metadata


def save_file_in_env(env, filename, metadata=None):
    if metadata:
        assert type(metadata) is FileMetadata
        metadata = dict(metadata._asdict())

    file_dict = env.get('files', {})
    file_dict[filename] = metadata
    env['files'] = file_dict


def read_response_to_file(response):
    # find the place we want to store files, and make sure it exists
    logger.info('Finding the place we want to store files')
    file_dir = settings.TMP_FILE_ROOT
    os.makedirs(file_dir, exist_ok=True)

    # random filename
    filename = uuid.uuid4().hex
    path = file_dir + filename

    # drains response.content and returns the name of the file it created
    filename = stream.stream_response_to_file(response, path=path)
    return filename


def calculate_hashes(file_path):
    file_hashes = {'md5': '', 'sha1': '', 'sha256': ''}
    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha256 = hashlib.sha256()
    try:
        with open(file_path, "rb") as file_object:
            for chunk in iter(lambda: file_object.read(4096), b''):
                md5.update(chunk)
                sha1.update(chunk)
                sha256.update(chunk)
        file_hashes['md5'] = md5.hexdigest()
        file_hashes['sha1'] = sha1.hexdigest()
        file_hashes['sha256'] = sha256.hexdigest()
    except IOError as e:
        logger.warn('Could not open file:{} :: {}'.format(file_path, str(e)))
    return file_hashes
